/*
 * Package: auto-pilot
 *
 * Erez Zadok <ezk@cs.sunysb.edu>
 * Charles P. Wright <cwright@cs.sunysb.edu>
 *
 * Copyright (c) 2001-2005 Stony Brook University
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/fcntl.h>
#include <asm/param.h>
#include <dirent.h>
#include <errno.h>
#include <regex.h>

#define HASHSIZE 101

#define DIFF 1
#define CAPTURE 2
#define SUMMARY 3
#define MATCH 4

int *exclude = NULL;
static struct ptime *phash[HASHSIZE];
static unsigned long threshold = 0;
static unsigned char *pattern;
unsigned long summary = 0;

int ismatch(const char *pat, const char *name) {
	int err;
	char errbuf[1024];
	static regex_t re;

	if (err = regcomp(&re, pat, REG_EXTENDED|REG_NOSUB)) {
		regerror(err, &re, errbuf, sizeof(errbuf));	
		fprintf(stderr, "regcomp(%s): %s\n", pat, errbuf);
		exit(1);
	}

	err = regexec(&re, name, 0, NULL, 0);
	regfree(&re);
	if (!err) {
		return 1;
	} else if (err == REG_NOMATCH) {
		return 0;
	} else {
		regerror(err, &re, errbuf, sizeof(errbuf));	
		fprintf(stderr, "regexec(%s, %s): %s\n", pat, name, errbuf);
		exit(1);
	}
}

struct ptime {
	int pt_pid;
	unsigned long pt_utime;
	unsigned long pt_stime;
	long pt_cutime;
	long pt_cstime;
	int pt_touch;
	struct ptime *pt_next;
	int pt_namelen;
	char pt_name[0];
};

void add_proc(int pid, const char *name, unsigned long utime, unsigned long stime, long cutime, long cstime) {
	struct ptime *current;

	current = (struct ptime *)malloc(sizeof(struct ptime) + strlen(name) + 1);	
	if (!current) {
		perror("malloc");
		exit(1);
	}

	current->pt_pid = pid;
	current->pt_namelen = strlen(name);
	strcpy(current->pt_name, name);
	current->pt_utime = utime;
	current->pt_stime = stime;
	current->pt_cutime = cutime;
	current->pt_cstime = cstime;
	current->pt_touch = 0;

	current->pt_next = phash[pid % HASHSIZE];
	phash[pid % HASHSIZE] = current;
}

void check_proc(int pid, const char *name, unsigned long utime, unsigned long stime, long cutime, long cstime) {
	struct ptime *current;

	current = phash[pid % HASHSIZE];
	while(current) {
		if (current->pt_pid == pid) {
			unsigned long udiff;
			unsigned long sdiff;
			long csdiff;
			long cudiff;
			if (!strcmp(current->pt_name, name)) {
				current->pt_touch = 1;
				csdiff = (cstime - current->pt_cstime);
				cudiff = (cutime - current->pt_cutime);
				sdiff = (stime - current->pt_stime);
				udiff = (utime - current->pt_utime);
				if ((udiff + sdiff + cudiff + csdiff) > threshold) {
					printf("[procdiff]\n");
					printf("type = ptime\n");
					printf("pid = %d\n", pid);
					printf("name = \"%s\"\n", name);
					printf("user = %.06f\n", (float)udiff / (float)HZ);
					printf("sys = %.06f\n", (float) sdiff/ (float)HZ);
					printf("cuser = %.06f\n", (float) cudiff/ (float)HZ);
					printf("csys = %.06f\n", (float)csdiff / (float)HZ );
					printf("\n");
				}
			} else {
				printf("[procdiff]\n");
				printf("type = change\n");
				printf("pid = %d\n", pid);
				printf("old = %s\n", current->pt_name);
				printf("new = %s\n", name);
				printf("\n");
			}
			return;
		}
		current = current->pt_next;
	}

	printf("[procdiff]\n");
	printf("type = start\n");
	printf("pid = %d\n", pid);
	printf("name = %s\n", name);
	printf("user = %.06f\n", utime / (float)HZ);
	printf("sys = %.06f\n", stime / (float)HZ);
	printf("cuser = %.06f\n", cutime / (float)HZ);
	printf("csys = %.06f\n", cstime / (float)HZ);
	printf("\n");
	

	return;
}

unsigned long diff_proc(int pid, const char *name, unsigned long utime, unsigned long stime, long cutime, long cstime) {
	struct ptime *current;
	unsigned long diff = 0;

	current = phash[pid % HASHSIZE];
	while(current) {
		if (current->pt_pid == pid) {
			if (!strcmp(current->pt_name, name)) {
				current->pt_touch = 1;
				diff += (cstime - current->pt_cstime);
				diff += (cutime - current->pt_cutime);
				diff += (stime - current->pt_stime);
				diff += (utime - current->pt_utime);
			} else {
				diff = utime + stime + cutime + cstime;
			}
			goto out;
		}
		current = current->pt_next;
	}

	/* We just started. */
	diff = utime + stime + cutime + cstime;

out:
	if (diff < threshold) {
		diff = 0;
	}
	return diff;
}


/* CPW: Look at procps. */
void stat2proc(char* S, unsigned long *utime, unsigned long *stime, long *cutime, long *cstime) {
    int num;

    char* tmp = strrchr(S, ')');	/* split into "PID (cmd" and "<rest>" */
    *tmp = '\0';			/* replace trailing ')' with NUL */
    num = sscanf(tmp + 2,			/* skip space after ')' too */
       "%*c "
       "%*d %*d %*d %*d %*d "
       "%*u %*u %*u %*u %*u %lu %lu " /* Last two are utime and stime */
       "%ld %ld %*d %*d %*d %*d " /* First two are cutime and cstime */
       "%*u %*u "
       "%*d "
       "%*u %*u %*u %*u %*u %*u "
       "%*s %*s %*s %*s " /* discard, no RT signals & Linux 2.1 used hex */
       "%*u %*u %*u %*d %*d",
       utime, stime, cutime, cstime);
}

int find_pid(int pid, int mode) {
	FILE *f = NULL;
	unsigned long utime;
	unsigned long stime;
	long cstime;
	long cutime;
	char path[4096];
	char name[4096];
	char temp[4096];
	int fd, n, i;

	snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);

	fd = open(path, O_RDONLY);
	if (fd == -1) {
		perror("open");
		return -1;
	}
	memset(name, '\0', sizeof(name));
	n = read(fd, name, sizeof(name));
	if (n == -1) {
		perror("read");
		return -1;
	}
	if (n && name[0]) {
		for (i = 0; i < n - 1; i++) {
			if (name[i] == '\0') {
				name[i] = ' ';
			}
		}
	}
	close(fd);

	snprintf(path, sizeof(path), "/proc/%d/stat", pid);
	f = fopen(path, "r");
	memset(temp, '\0', sizeof(temp));
	fgets(temp, sizeof(temp), f);
	temp[sizeof(temp) - 1] = '\0';
	fclose(f);


	stat2proc(temp, &utime, &stime, &cutime, &cstime);

	snprintf(path, sizeof(path), "/proc/%d/stat", pid);
	f = fopen(path, "r");
	if (f) {
		memset(temp, '\0', sizeof(temp));
		fgets(temp, sizeof(temp), f);
		fclose(f);
	} else {
		perror("fopen");
		return -1;
	}

	if (name[0] == '\0') {
		char *p = strchr(temp, '(');;
		if (p) {
			char *q = strchr(p, ')');
			p++;
			*q = '\0';
			strcpy(name, p);
		}
	}

	switch(mode) {
	case DIFF:
		check_proc(pid, name, utime, stime, cutime, cstime);
		break;
	case CAPTURE:
		add_proc(pid, name, utime, stime, cutime, cstime);
		break;
	case SUMMARY:
		if (!pattern || ismatch(pattern, name)) {
			summary += diff_proc(pid, name, utime, stime, cutime, cstime);	
		}
		break;
	default:
		fprintf(stderr, "BUG: Invalid mode.\n");
		exit(1);
	}

	return 0;
}

void print_untouch() {
	int i;
	struct ptime *current;

	for (i = 0; i < sizeof(phash) / sizeof(struct ptime *); i++) {
		current = phash[i];	
		while(current) {
			if (!current->pt_touch) {
				printf("[procdiff]\n");
				printf("type = exit\n");
				printf("name = %d\n", current->pt_pid);
				printf("name = %s\n", current->pt_name);
				printf("\n");
			}
			current = current->pt_next;
		}
	}
}

void writeproc(const char *file) {
	int fd;
	int i;
	struct ptime *current;

	fd = open(file, O_WRONLY|O_CREAT|O_EXCL, 0600);

	if (fd == -1) {
		perror("open");
		exit(1);
	}

	for (i = 0; i < sizeof(phash) / sizeof(struct ptime *); i++) {
		current = phash[i];	
		while(current) {
			write(fd, current, sizeof(*current) + current->pt_namelen + 1);
			current = current->pt_next;
		}
	}

	close(fd);
}

void readproc(const char *file) {
	int fd;
	int ret;
	struct ptime current;
	char name[PATH_MAX];

	fd = open(file, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "open(%s): %s\n", file, strerror(errno));
		exit(1);
	}

	while((ret = read(fd, &current, sizeof(current)) == sizeof(current))) {
		if (read(fd, name, current.pt_namelen + 1) != (current.pt_namelen + 1)) {
			perror("read");
		}	
		add_proc(current.pt_pid, name, current.pt_utime, current.pt_stime, current.pt_cutime, current.pt_cstime);
	}
	if (ret) {
		perror("read");
	}

	close(fd);
}

void usage() {
	fprintf(stderr, "procdiff -m [capture|diff|summary|match] -f file -t [threshold msec]\n");
	exit(1);
}

int main(int argc, char *argv[]) {
	DIR *dir;
	char *statefile = NULL;
	struct dirent *ent;
	int delay = 0;
	int mode;
	int serialize = 0;
	int deserialize = 0;
	char *end;
	int c;

	while ((c = getopt(argc, argv, "m:f:t:p:")) != -1) {
	switch(c) {
	case 'm':
		if (!optarg) usage();
		if (!strcmp(optarg, "capture")) {
			mode = CAPTURE;	
			serialize = 1;
		} else if (!strcmp(optarg, "diff")) {
			mode = DIFF;
			deserialize = 1;
		} else if (!strcmp(optarg, "summary")) {
			mode = SUMMARY;
			deserialize = 1;
		} else if (!strcmp(optarg, "match")) {
			mode = MATCH;
			deserialize = 1;
		} else {
			usage();
		}
		break;
	case 'f':
		if (!optarg) usage();
		statefile = strdup(optarg);
		break;
	case 't':
		if (!optarg) usage();
		threshold = strtol(optarg, &end, 0);
		if (*end) usage();
		threshold *= HZ;
		threshold /= 1000;
		break;
	case 'p':
		if (!optarg) usage();
		pattern = strdup(optarg);
		break;
	}
	}

	if (!mode) {
		usage();
	}
	if (!statefile) {
		usage();
	}

	if (deserialize) {
		readproc(statefile);
	}

	dir = opendir("/proc");
	if (!dir) {
		perror("opendir");
		exit(1);
	}
	while((ent = readdir(dir))) {
		if (ent->d_name[0] >= '0' && ent->d_name[0] <= '9') {
			char *end;
			strtol(ent->d_name, &end, 0);
			if (*end) {
				continue;
			}
			find_pid(atoi(ent->d_name), mode);
		}
	}
	closedir(dir);

	switch(mode) {
	case DIFF:
		print_untouch();
		break;
	case SUMMARY:
		if (pattern) {
			char *prettypattern = strdup(pattern);
			int i = 0;
			while(*pattern && !isalnum(*pattern)) {
				pattern++;
			}
			while (isalnum(*pattern)) {
				prettypattern[i++] = *pattern++;
			}
			prettypattern[i] = '\0';
			printf("%s = %.06lf\n", prettypattern, (double)summary / (double)HZ);
			free(prettypattern);
		} else {
			printf("procdiff = %.06lf\n", (double)summary / (double)HZ);
		}
		break;
	}

	if (serialize) {
		writeproc(statefile);
		/* Do serialization. */
	}	


	exit(0);
}
